import re
import zlib
import base64
import calendar
import itertools
from six import iterkeys, iteritems, PY3

if PY3:
    long = int


def dt_epoch_msecs(value):
    """
    Calculate miliseconds since epoch start for python datetimes.
    """
    return long(calendar.timegm(value.timetuple())) * 1000


def np_dt_epoch_msec(value):
    """
    Calculate miliseconds since epoch start for numpy datetimes.
    """
    return value.astype(long) / 1000


def encode_string(string, level=9):
    """
    Compresses and base64 encodes the supplied string
    :param string: String to encode and compress
    :param level: Compression level
    :return: Compressed and encoded string
    """
    return base64.b64encode(zlib.compress(string.encode('utf8'), level)[2:-4])


def camelcase(value):
    """
    Converts 'under_score_string' -> 'underScoreString'

    :param value: Underscore separated string
    :return: CamelCased string
    """
    rest = value.split("_")
    return rest[0] + "".join(word.title() for word in rest[1:])


def underscorecase(camelcased):
    """
    Converts 'underScoreString' -> 'under_score_string'

    :param value: CamelCased string
    :return: Underscore separated string
    """
    return re.sub('([A-Z]+)', r'_\1', camelcased).lower()


def cfg_to_prop_string(cfg, key_transform=lambda k: k, value_transform=lambda v: v, separator=";"):
    """
    Convert the config object to a property string. Useful for constructing CSS and javascript
    object init strings.

    Underscores are replaced with dashes and values are converted to lower case.
    """
    return separator.join(["%s:%s" % (key_transform(key), value_transform(value)) for key, value in iteritems(cfg)])


def cfg_to_css_string(cfg):
    return cfg_to_prop_string(cfg, lambda k: k.replace("_", "-"), lambda v: str(v).lower())


class Cfg(dict):
    """
    Provides plumbing for inheritable block parameters.
    """

    def inherit(self, parent):
        """
        Inherit all parent settings that the current config does not define.
        """
        return self.__class__(Cfg._mergedicts(self, parent, False))

    def inherit_many(self, *args, **kwargs):
        """
        Inherit many settings at once.
        """
        inherit_collapsed = self._collapse_args(args, kwargs)
        return self.inherit(inherit_collapsed)

    def override(self, parent):
        """
        Override all settings specified in the overrides.
        """
        return self.__class__(Cfg._mergedicts(self, parent, True))

    def override_many(self, *args, **kwargs):
        """
        Override many settings at once.
        """
        override_collapsed = self._collapse_args(args, kwargs)
        return self.override(override_collapsed)

    @staticmethod
    def _collapse_args(args, kwargs):
        inherit_agg = Cfg(kwargs)
        for arg in args:
            # If we got a type as configuration, instantiate it
            if isinstance(arg, type):
                arg = arg()

            inherit_agg = inherit_agg.inherit(arg)
        return inherit_agg

    @staticmethod
    def _mergedicts(dict1, dict2, take_second):
        for k in set(itertools.chain(iterkeys(dict1), iterkeys(dict2))):
            if k in dict1 and k in dict2:
                v1 = dict1[k]
                v2 = dict2[k]
                if isinstance(v1, dict) and isinstance(v2, dict):
                    # The merged sub-dict type should be the operand type based on take_second
                    ctor = v2.__class__ if take_second else v1.__class__
                    yield (k, ctor(Cfg._mergedicts(v1, v2, take_second)))
                else:
                    yield (k, v2 if take_second else v1)
            elif k in dict1:
                yield (k, dict1[k])
            else:
                yield (k, dict2[k])

    def __getattr__(self, item):
        return self[item]

    def __setattr__(self, name, value):
        self[name] = value



from abc import ABCMeta, abstractmethod
from io import open
import logging
import os
import subprocess
import sys
from pybloqs.config import user_config

logger = logging.getLogger(__name__)

PORTRAIT = 'Portrait'
LANDSCAPE = 'Landscape'
A4 = 'A4'


class HTMLConverter(object):
    """
    Definition of interface for HTML to X converters
    """
    __metaclass__ = ABCMeta

    def get_executable(self, command_name):
        """Look for executable file name in VENV_NAME/bin/ otherwise use plain command name and hope it is in PATH."""
        py_bin_dir = os.path.join(sys.exec_prefix, 'bin')
        local_bin = os.path.join(py_bin_dir, command_name)
        if os.path.isfile(local_bin):
            command = local_bin
        else:
            command = command_name
        return command

    def run_command(self, cmd):
        """Run cmd as subprocess and return stdout and stderr."""
        logger.info('Running external application: {}'.format(cmd))
        proc = subprocess.Popen(cmd)

        # Wait for the process to exit
        output, errors = proc.communicate()

        if proc.returncode != 0:
            raise ValueError("{} returned:\n stdout:{}\n stderr:{}".format(cmd, output, errors))
        else:
            logger.info('Returned:\n stdout: {}\n stderr:{}'.format(output, errors))
        return output, errors

    @staticmethod
    def write_html_to_tempfile(block, content):
        name = block._id[:user_config["id_precision"]] + ".html"
        tempdir = user_config["tmp_html_dir"]
        html_filename = os.path.join(tempdir, name)
        with open(html_filename, "w", encoding='utf-8') as f:
            f.write(content)
        return html_filename

    @staticmethod
    def remove_temporary_files(temp_files):
        if user_config.get('remove_temp_files', True):
            for f in temp_files:
                try:
                    os.remove(f)
                except (IOError, OSError):
                    logger.exception('Failed to remove a temporary file: {}.'.format(f))

    @abstractmethod
    def htmlconv(self, input_file, output_file, header_filename=None, header_spacing=None,
                 footer_filename=None, footer_spacing=None, zoom=1, pdf_page_size=A4, orientation=PORTRAIT, **kwargs):
        pass



from io import open
import os
import tempfile
import uuid
import webbrowser

from six import BytesIO

from pybloqs.config import user_config
from pybloqs.email import send_html_report
from pybloqs.html import root, append_to, render, js_elem, id_generator
import pybloqs.htmlconv as htmlconv

from pybloqs.static import DependencyTracker, Css, script_inflate, script_block_core, register_interactive
from pybloqs.util import Cfg, cfg_to_css_string
from six.moves.urllib.parse import urljoin


default_css_main = Css(os.path.join("css", "pybloqs_default", "main"))
register_interactive(default_css_main)


class BaseBlock(object):
    """
    Base class for all blocks. Provides infrastructure for rendering the block
    in an IPython Notebook or saving it to disk in HTML, PDF, PNG or JPG format.
    """
    container_tag = "div"
    resource_deps = []

    def __init__(self, title=None, title_level=3, title_wrap=False,
                 width=None, height=None, inherit_cfg=True,
                 styles=None, classes=(), anchor=None, **kwargs):
        self._settings = Cfg(title=title,
                             title_level=title_level,
                             title_wrap=title_wrap,
                             cascading_cfg=Cfg(**kwargs).override(styles or Cfg()),
                             default_cfg=Cfg(),
                             inherit_cfg=inherit_cfg,
                             width=width,
                             height=height,
                             classes=["pybloqs"] + ([classes] if isinstance(classes, str) else list(classes)))
        # Anchor should not be inherited, so keep outside of Cfg
        self._anchor = anchor
        self._id = uuid.uuid4().hex

    def render_html(self, pretty=True, static_output=False, header_block=None, footer_block=None):
        """Returns html output of the block
        :param pretty: Toggles pretty printing of the resulting HTML. Not applicable for non-HTML output.
        :param static_output: Passed down to _write_block. Will render static version of blocks which support this.
        :param header_block: If not None, header is inlined into a HTML body as table.
        :param footer_block: If not None, header is inlined into a HTML body as table.
        :return html-code of the block
        """
        # Render the contents
        html = root("html", doctype="html")
        head = append_to(html, "head")
        append_to(head, "meta", charset='utf-8')

        body = append_to(html, "body")

        # Make sure that the main style sheet is always included
        resource_deps = DependencyTracker(default_css_main)

        # If header or footer are passed into this function, inline them in the following structure:
        #
        # <body>
        # <table>
        #    <thead><tr><td>Header html</td></tr></thead>
        #    <tfoot><tr><td>Footer html</td></tr></tfoot>
        #    <tbody><tr><td>Body html</td></tr></tbody>
        # </table>
        # </body>
        if header_block is not None or footer_block is not None:
            content_table = append_to(body, "table")
            if header_block is not None:
                header_thead = append_to(content_table, "thead")
                header_tr = append_to(header_thead, "tr")
                header_td = append_to(header_tr, "th")
                header_block._write_block(header_td, Cfg(), id_generator(), resource_deps=resource_deps,
                                          static_output=static_output)

            if footer_block is not None:
                footer_tfoot = append_to(content_table, "tfoot", id='footer')
                footer_tr = append_to(footer_tfoot, "tr")
                footer_td = append_to(footer_tr, "td")
                footer_block._write_block(footer_td, Cfg(), id_generator(), resource_deps=resource_deps,
                                          static_output=static_output)

            body_tbody = append_to(content_table, "tbody")
            body_tr = append_to(body_tbody, "tr")
            body_td = append_to(body_tr, "td")
            self._write_block(body_td, Cfg(), id_generator(), resource_deps=resource_deps, static_output=static_output)
        else:
            self._write_block(body, Cfg(), id_generator(), resource_deps=resource_deps, static_output=static_output)

        script_inflate.write(head)
        script_block_core.write(head)

        if static_output:
            # Add the load wait poller if there are any JS resources
            js_elem(body, "var loadWaitPoller=runWaitPoller();")

        # Write out resources
        for res in resource_deps:
            res.write(head)

        # Render the whole document (the parent of the html tag)
        content = render(html.parent, pretty=pretty)
        return content

    def save(self, filename=None, fmt=None, pdf_zoom=1, pdf_page_size=htmlconv.html_converter.A4, pdf_auto_shrink=True,
             orientation=htmlconv.html_converter.PORTRAIT, header_block=None, header_spacing=5, footer_block=None,
             footer_spacing=5, **kwargs):
        """
        Render and save the block. Depending on whether the filename or the format is
        provided, the content will either be written out to a file or returned as a string.

        :param filename: Format will be based on the file extension.
                         The following formats are supported:
                         - HTML
                         - PDF
                         - PNG
                         - JPG
        :param fmt: Specifies the format of a temporary output file. When supplied, the filename
                    parameter must be omitted.
        :param pdf_zoom: The zooming to apply when rendering the page.
        :param pdf_page_size: The page size to use when rendering the page to PDF.
        :param pdf_auto_shrink: Toggles auto-shrinking content to fit the desired page size (wkhtmltopdf only)
        :param orientation: Either html_converter.PORTRAIT or html_converter.LANDSCAPE
        :param header_block: Block to be used as header (and repeated on every page). Only used for PDF output.
        :param header_spacing: Size of header block. Numbers are in mm. HTML sizes (e.g. '5cm') in chrome_headless only.
        :param footer_block: Block to be used as footer (and repeated on every page). Only used for PDF output.
        :param footer_spacing: Size of header block. Numbers are in mm. HTML sizes (e.g. '5cm') in chrome_headless only.
        :return: html filename
        """
        # Ensure that exactly one of filename or fmt is provided
        if filename is None and fmt is None:
            raise ValueError("One of `filename` or `fmt` must be provided.")

        tempdir = user_config["tmp_html_dir"]

        if filename:
            _, fmt_from_name = os.path.splitext(filename)
            # Exclude the dot from the extension, gosh darn it!
            fmt_from_name = fmt_from_name[1:]
            if fmt is None:
                if fmt_from_name == '':
                    raise ValueError('If fmt is not specified, filename must contain extension')
                fmt = fmt_from_name
            else:
                if fmt != fmt_from_name:
                    filename += '.' + fmt
        else:
            name = self._id[:user_config["id_precision"]] + "." + fmt
            filename = os.path.join(tempdir, name)

        # Force extension to be lower case so format checks are easier later
        fmt = fmt.lower()

        is_html = "htm" in fmt

        if is_html:
            content = self.render_html(static_output=False, header_block=header_block, footer_block=footer_block)
            with open(filename, "w", encoding='utf-8') as f:
                f.write(content)
        else:
            converter = htmlconv.get_converter(fmt)
            converter.htmlconv(self, filename,
                               header_block=header_block, header_spacing=header_spacing,
                               footer_block=footer_block, footer_spacing=footer_spacing,
                               pdf_page_size=pdf_page_size,
                               orientation=orientation,
                               pdf_auto_shrink=pdf_auto_shrink,
                               pdf_zoom=pdf_zoom,
                               **kwargs)
        return filename

    def publish(self, name, *args, **kwargs):
        """
        Publish the block so that others can access it.

        :param name: Name to publish under. Can be a filename or a relative path.
        :param args: Arguments to pass to `Block.save`.
        :param kwargs: Keyword arguments to pass to `Block.save`.
        :return: Path to the published block file.
        """
        full_path = os.path.join(user_config["public_dir"], name)
        full_path = os.path.expanduser(full_path)

        base_dir = os.path.dirname(full_path)

        try:
            os.makedirs(base_dir)
        except OSError:
            pass  # Directory already exists

        self.save(full_path, * args, **kwargs)

        return full_path

    def show(self, fmt="html", header_block=None, footer_block=None):
        """
        Show the block in a browser.

        :param fmt: The format of the saved block. Supports the same output as `Block.save`
        :return: Path to the block file.
        """
        file_name = self._id[:user_config["id_precision"]] + "." + fmt
        file_path = self.publish(os.path.expanduser(os.path.join(user_config["tmp_html_dir"], file_name)),
                                 header_block=header_block, footer_block=footer_block)

        try:
            url_base = user_config["public_dir"]
        except KeyError:
            path = os.path.expanduser(file_path)
        else:
            path = urljoin(url_base, os.path.expanduser(user_config["tmp_html_dir"] + "/" + file_name))

        webbrowser.open_new_tab(path)

        return path

    def email(self, title="", recipients=(user_config["user_email_address"],),
              header_block=None, footer_block=None,
              from_address=None, cc=None, bcc=None, attachments=None,
              convert_to_ascii=True, **kwargs):
        """
        Send the rendered blocks as email. Each output format chosen will be added as an
        attachment.

        :param title: title of the email
        :param recipients: recipient of the email
        :param fmt: One or more output formats that should be included as attachments.
                    The following formats are supported:
                    - HTML
                    - PDF
                    - PNG
                    - JPG
        :param body_block: The block to use as the email body. The default behavior is
                          to use the current block.
        :param from_address: sender of the message. Defaults to user name.
            Can be overwritten in .pybloqs.cfg with yaml format: 'user_email_address: a@b.com'
        :param cc: cc recipient
        :param bcc: bcc recipient
        :param convert_to_ascii: bool to control convertion of html email to ascii or to leave in current format
        :param kwargs: Optional arguments to pass to `Block.render_html()`
        """
        if from_address is None:
            from_address = user_config["user_email_address"]

        # The email body needs to be static without any dynamic elements.
        email_html = self.render_html(header_block=header_block, footer_block=footer_block, **kwargs)

        send_html_report(email_html, recipients, subject=title, attachments=attachments,
                         From=from_address, Cc=cc, Bcc=bcc, convert_to_ascii=convert_to_ascii)

    def to_static(self):
        return self._visit(lambda block: block._to_static())

    def _to_static(self):
        """
        Subclasses can override this method to provide a static content version.
        """
        return self

    def _visit(self, visitor):
        """
        Calls the supplied visitor function on this block and any sub-blocks
        :param visitor: Visitor function
        :return: Return value of the visitor
        """
        return visitor(self)

    def _provide_default_cfg(self, defaults):
        """
        Makes the supplied config to be part of the defaults for the block.
        :param defaults: The default parameters that should be inherited.
        """
        self._settings.default_cfg = self._settings.default_cfg.inherit(defaults)

    def _combine_parent_cfg(self, parent_cfg):
        """from pybloqs.config import user_config
        Combine the supplied parent and the current Block's config.

        :param parent_cfg: Parent config to inherit from.
        :return: Combined config.
        """
        # Combine parameters only if inheritance is turned on
        if self._settings.inherit_cfg:
            actual_cfg = self._settings.cascading_cfg.inherit(parent_cfg)
        else:
            actual_cfg = self._settings.cascading_cfg

        # Any undefined settings will use the defaults
        actual_cfg = actual_cfg.inherit(self._settings.default_cfg)

        return actual_cfg

    def _get_styles_string(self, styles_cfg):
        """
        Converts the styles configuration to a CSS styles string.

        :param styles_cfg: The configuration object to convert.
        :return: CSS string
        """
        sizing_cfg = Cfg()

        if self._settings.width is not None:
            sizing_cfg["width"] = self._settings.width

        if self._settings.height is not None:
            sizing_cfg["height"] = self._settings.height

        # Replace `_` with `-` and make values lowercase to get valid CSS names
        return cfg_to_css_string(styles_cfg.override(sizing_cfg))

    def _write_block(self, parent, parent_cfg, id_gen, resource_deps=None, static_output=False):
        """
        Writes out the block into the supplied stream, inheriting the parent_parameters.

        :param parent: Parent element
        :param parent_cfg: Parent parameters to inherit.
        :param id_gen: Unique ID generator.
        :param resource_deps: Object used to register resource dependencies.
        :param static_output: A value of True signals to blocks that the final output will
                              be a static format. Certain dynamic content will render with
                              alternate options.

        """
        if resource_deps is not None:
            for res in self.resource_deps:
                resource_deps.add(res)

        actual_cfg = self._combine_parent_cfg(parent_cfg)

        if self.container_tag is not None:
            container = append_to(parent, self.container_tag)
            self._write_container_attrs(container, actual_cfg)
        else:
            container = parent

        self._write_anchor(container)
        self._write_title(container)
        self._write_contents(container, actual_cfg, id_gen, resource_deps=resource_deps, static_output=static_output)

    def _write_container_attrs(self, container, actual_cfg):
        """
        Writes out the container attributes (styles, class, etc...).
        Note that this method will only be called if the container tag is not `None`.

        :param container: Container element.
        :param actual_cfg: Actual parameters to use.
        """
        styles = self._get_styles_string(actual_cfg)
        if len(styles) > 0:
            container["style"] = styles

        container["class"] = self._settings.classes

    def _write_title(self, container):
        """
        Write out the title (if there is any).

        :param container: Container element.
        """
        if self._settings.title is not None and (self._settings.title != ""):
            title = append_to(container, "H%s" % self._settings.title_level,
                              style="white-space: %s" % ("normal" if self._settings.title_wrap else "nowrap"))
            title.string = self._settings.title

    def _write_anchor(self, container):
        """
        Write HTML anchor for linking within page

        :param container: Container element.
        """
        if self._anchor is not None:
            append_to(container, "a", name=self._anchor)

    def _write_contents(self, container, actual_cfg, id_gen, resource_deps=None, static_output=None):
        """
        Write out the actual contents of the block. Deriving classes must override
        this method.

        :param container: Container element.
        :param actual_cfg: Actual parameters to use.
        :param id_gen: Unique ID generator.
        :param resource_deps: Object used to register resource dependencies.
        :param static_output: A value of True signals to blocks that the final output will
                              be a static format. Certain dynamic content will render with
                              alternate options.
        """
        raise NotImplementedError("_write_contents")

    def _repr_html_(self, *_):
        """
        Function required to support interactive IPython plopping and plotting.

        Should not be used directly.

        :return: Data to be displayed
        """
        return self.data.decode()

    @property
    def data(self):
        """
        Function required to support interactive IPython plotting.

        Should not be used directly.

        :return: Data to be displayed
        """
        container = root("div")
        self._write_block(container, Cfg(), id_generator())

        # Write children into the output
        output = BytesIO()

        for child in container.children:
            output.write(render(child).encode('utf-8'))

        return output.getvalue()


class HRule(BaseBlock):
    """
    Draws a horizontal divider line.
    """

    def _write_block(self, parent, *args, **kwargs):
        # Add a `hr` element to the parent
        append_to(parent, "hr")



from pybloqs.plot import (
    Plot, Options3d, Chart, Legend, XAxis, YAxis, Expr, ColorAxis, Tooltip, Heatmap as HeatmapPlot, Scatter,
    Title)


class Heatmap(Plot):
    def __init__(self, data, *args, **kwargs):
        super(Heatmap, self).__init__(data, HeatmapPlot(), flatten=True, *args, **kwargs)

        self._chart_cfg = self._chart_cfg.inherit_many(
            YAxis(title=None, reversed=True), XAxis(opposite=True),
            ColorAxis(min_color="#ffffff", max_color="#3060cf"))


class Corr(Heatmap):
    def __init__(self, data, *args, **kwargs):
        super(Corr, self).__init__(data, *args, **kwargs)

        self._chart_cfg = self._chart_cfg.inherit_many(
            ColorAxis(min=0, max=1),
            Tooltip(formatter=Expr("""\
                function() {
                return '<b>' + this.series.xAxis.categories[this.point.x] + ' - '
                       + this.series.yAxis.categories[this.point.y] + '</b>: '
                       + ((this.point.value * 100) | 0) + '%';
                }
                """)))


class Surface(Plot):
    def __init__(self, data, *args, **kwargs):
        options = Options3d(enabled=True,
                            alpha=kwargs.pop("alpha", 20),
                            beta=kwargs.pop("beta", 0),
                            depth=kwargs.pop("depth", 150),
                            view_distance=kwargs.pop("view_distance", 10))

        super(Surface, self).__init__(data, Scatter(), flatten=True, switch_zy=True, *args, **kwargs)

        self._chart_cfg = self._chart_cfg.inherit_many(Chart(options),
                                                       Legend(enabled=False),
                                                       YAxis(Title(text=None))).override_many(Chart(zoom_type=None))

    def _write_plot_postprocess(self, chart_buf):
        chart_buf.write("pybloqsSurfaceRotate(chart);")



import base64
from six import StringIO, BytesIO
import struct
from contextlib import contextmanager

from matplotlib.artist import Artist
from matplotlib.figure import Figure

import matplotlib.pyplot as plt
from pybloqs.block.base import BaseBlock
from pybloqs.block.convenience import add_block_types
from pybloqs.html import append_to, parse
from pybloqs.static import JScript, Css
from pybloqs.util import cfg_to_css_string

try:
    from plotly.graph_objs import Figure as PlotlyFigure
    import plotly.offline as po
    _PLOTLY_AVAILABLE = True
except ImportError:
    _PLOTLY_AVAILABLE = False

try:
    from bokeh.resources import CSSResources, JSResources
    from bokeh.plotting.figure import Figure as BokehFigure
    from bokeh.embed.standalone import components
    _BOKEH_AVAILABLE = True
except ImportError:
    _BOKEH_AVAILABLE = False


_MIME_TYPES = {
    "png": "png",
    "svg": "svg+xml"
}

_PLOT_FORMAT = "png"
_PLOT_MIME_TYPE = _MIME_TYPES[_PLOT_FORMAT]
_PLOT_DPI = 100


@contextmanager
def plot_format(plot_format=None, dpi=None):
    """
    Temporarily set the plot formatting settings

    :param plot_format: The plot format (e.g 'png')
    :type plot_format:  str
    :param dpi:         The DPI of the plots
    :type dpi:          int
    """
    old = get_plot_format()
    set_plot_format(plot_format, dpi)
    yield
    set_plot_format(*old)


def get_plot_format():
    """
    Get the current plot format parameters

    :return: tuple of format and dpi
    """
    return _PLOT_FORMAT, _PLOT_DPI


def set_plot_format(plot_format=None, plot_dpi=None):
    """
    Overwrite the current plot format settings

    :param plot_format: The plot format (e.g. 'png')
    :type  plot_format: str
    :param plot_dpi:    The DPI of the plots
    :type  plot_dpi:    int
    """
    global _PLOT_FORMAT
    global _PLOT_MIME_TYPE
    global _PLOT_DPI
    if plot_format is not None:
        _PLOT_FORMAT = plot_format
        _PLOT_MIME_TYPE = _MIME_TYPES[plot_format]

    if plot_dpi is not None:
        _PLOT_DPI = plot_dpi


class ImgBlock(BaseBlock):

    def __init__(self, data, mime_type="png", width=None, height=None, img_style=None, **kwargs):
        """
        Create a block containing an image. The dimensions can be sniffed from GIF
        and PNG data, for other formats, the `width` and `height` parameters can be
        used to specify the size.

        :param data: Either a block to convert to an image, or raw image data.
        :mime_type: The mime type of the image (e.g. "png" or "svg+xml")
        :param width: Image width override as string, e.g. "50px". If unspecified, the actual
                      image width is used.
        :param width: Image height override as string, e.g. "50px". If unspecified, the actual
                      image height is used.
        :param kwargs: Optional styling arguments. The `style` keyword argument has special
                       meaning in that it allows styling to be grouped as one argument.
                       It is also useful in case a styling parameter name clashes with a standard
                       block parameter.
        """
        # Wrapping a block in an image will render it
        if hasattr(data, "_write_block"):
            img_file = data.save(fmt=mime_type)
            data = open(img_file, "rb").read()

        self._img_data = base64.b64encode(data)
        self._mime_type = mime_type

        if width is None and height is None:
            if mime_type.lower() == "png":
                if struct.unpack('ccc', data[1:4]) != (b'P', b'N', b'G'):
                    raise ValueError('Image type is not png and does not match mime type')
                x, y = struct.unpack(">ii", data[16:24])
            elif mime_type.lower() == "gif":
                x, y = struct.unpack("<HH", data[6:10])
            else:
                raise ValueError("Can't determine image dimensions for mime type %s" % mime_type)

            width, height = ("%spx" % x, "%spx" % y)

        if img_style is None:
            img_styles = {}
        else:
            img_styles = img_style

        if width is not None:
            img_styles["width"] = width
        if height is not None:
            img_styles["height"] = height

        self._img_styles = img_styles

        super(ImgBlock, self).__init__(**kwargs)

    def _write_contents(self, container, *args, **kwargs):
        src = StringIO()
        mime = "data:image/{};base64,".format(self._mime_type)
        src.write(mime)
        src.write(self._img_data.decode())

        img = append_to(container, "img", src=src.getvalue())

        if len(self._img_styles) > 0:
            img["style"] = cfg_to_css_string(self._img_styles)

    @staticmethod
    def from_file(img_file, **kwargs):
        """
        Load an image block from a file.

        :param img_file: File path or file-like object.
        :param kwargs: Arguments to pass to the `ImgBlock` constructor.
        :return: ImgBlock instance.
        """
        close_file = False

        if isinstance(img_file, str):
            img_file = open(img_file, "rb")
            close_file = True

        try:
            return ImgBlock(img_file.read(), **kwargs)
        finally:
            # Close the file in case it was opened in this function
            if close_file:
                img_file.close()


class PlotBlock(ImgBlock):

    def __init__(self, plot, close_plot=True, bbox_inches="tight",
                 width=None, height=None, **kwargs):
        """
        Create a block containing a matplotlib figure

        :param plot: A matplotlib figure, axes or artist object.
        :close_plot: Optional (default=True). Set to True to close the plot after it is
                     captured into an image and avoid lingering plot windows.
        :bbox_inches: Optional bounding box parameter for 'figure.savefig'.
        :param kwargs: Optional styling arguments. The `style` keyword argument has special
                       meaning in that it allows styling to be grouped as one argument.
                       It is also useful in case a styling parameter name clashes with a standard
                       block parameter.
        """
        if not isinstance(plot, Artist):
            raise ValueError('PlotBlock contents must be matplotlib Artist')

        if isinstance(plot, Figure):
            figure = plot
        elif isinstance(plot, Artist):
            figure = plot.get_figure()
        else:
            raise ValueError("Unexpected plot object type %s", type(plot))

        img_data = BytesIO()

        legends = []

        for ax in figure.get_axes():
            legend = ax.get_legend()
            if legend is not None:
                # Patch Legend get_window_extent since there seems to be a bug where
                # it is passed an unexpected renderer instance.
                _orig_get_window_extent = legend.get_window_extent

                def _patched_get_window_extent(*_):
                    return _orig_get_window_extent()

                legend.get_window_extent = _patched_get_window_extent
                legends.append(legend)

        if len(figure.axes) == 0:
            # empty plot, disable bbox_inches to that savefig still works
            bbox_inches = None

        figure.savefig(img_data, dpi=_PLOT_DPI, format=_PLOT_FORMAT,
                       bbox_extra_artists=legends, bbox_inches=bbox_inches)

        plt_width, plt_height = figure.get_size_inches()

        width = width or "{:0.3f}in".format(plt_width)
        height = height or "{:0.3f}in".format(plt_height)

        if close_plot:
            plt.close(figure)

        super(PlotBlock, self).__init__(img_data.getvalue(),
                                        _PLOT_MIME_TYPE,
                                        width=width,
                                        height=height,
                                        **kwargs)

    def _to_static(self):
        # Convert to a basic image block in case we contain 'dynamic' svg content
        return ImgBlock(self) if self._mime_type == "svg" else super(PlotBlock, self)._to_static()


class PlotlyPlotBlock(BaseBlock):

    def __init__(self, contents, plotly_kwargs=None, **kwargs):
        """
        Writes out the content as raw text or HTML.

        :param contents: Plotly graphics object figure.
        :param plotly_kwargs: Kwargs that are passed to plotly plot function.
        :param kwargs: Optional styling arguments. The `style` keyword argument has special
                       meaning in that it allows styling to be grouped as one argument.
                       It is also useful in case a styling parameter name clashes with a standard
                       block parameter.
        """
        self.resource_deps = [JScript(script_string=po.offline.get_plotlyjs(), name='plotly')]

        super(PlotlyPlotBlock, self).__init__(**kwargs)

        if not isinstance(contents, PlotlyFigure):
            raise ValueError("Expected plotly.graph_objs.graph_objs.Figure type but got %s", type(contents))

        plotly_kwargs = plotly_kwargs or {}
        prefix = "<script>if (typeof require !== 'undefined') {var Plotly=require('plotly')}</script>"
        self._contents = prefix + po.plot(contents, include_plotlyjs=False, output_type='div', **plotly_kwargs)

    def _write_contents(self, container, *args, **kwargs):
        container.append(parse(self._contents))


class BokehPlotBlock(BaseBlock):

    def __init__(self, contents, **kwargs):
        """
        Writes out the content as raw text or HTML.

        :param contents: Bokeh plotting figure.
        :param kwargs: Optional styling arguments. The `style` keyword argument has special
                       meaning in that it allows styling to be grouped as one argument.
                       It is also useful in case a styling parameter name clashes with a standard
                       block parameter.
        """
        self.resource_deps = [JScript(script_string=s, name='bokeh_js') for s in JSResources().js_raw]
        self.resource_deps += [Css(css_string=s, name='bokeh_css') for s in CSSResources().css_raw]

        super(BokehPlotBlock, self).__init__(**kwargs)

        if not isinstance(contents, BokehFigure):
            raise ValueError("Expected bokeh.plotting.figure.Figure type but got %s", type(contents))

        script, div = components(contents)
        self._contents = script + div

    def _write_contents(self, container, *args, **kwargs):
        container.append(parse(self._contents))


add_block_types(Artist, PlotBlock)
# If Plotly or Bokeh are not installed skip registration
if _PLOTLY_AVAILABLE:
    add_block_types(PlotlyFigure, PlotlyPlotBlock)

if _BOKEH_AVAILABLE:
    add_block_types(BokehFigure, BokehPlotBlock)



from pybloqs.htmlconv.wkhtmltox import WkhtmltopdfConverter, WkhtmltoimageConverter
from pybloqs.config import user_config
from pybloqs.htmlconv.chrome_headless import ChromeHeadlessConverter

PDF_CONVERTERS = {
    'wkhtmltopdf': WkhtmltopdfConverter,
    'chrome_headless': ChromeHeadlessConverter,
}

IMAGE_CONVERTERS = {
    "wkhtmltoimage": WkhtmltoimageConverter,
}


def get_converter(file_type, **kwargs):
    """Parse the config and return the right backend for converting HTML to other formats."""
    if file_type.lower() == 'pdf':
        converter_name = user_config['pdf_converter']
        return PDF_CONVERTERS[converter_name](**kwargs)
    elif file_type.lower() in ['png', 'svg', 'jpg']:
        converter_name = user_config['image_converter']
        return IMAGE_CONVERTERS[converter_name](**kwargs)
    else:
        raise ValueError('No converter defined for file type: {}'.format(file_type))



import os

from pybloqs.htmlconv.html_converter import HTMLConverter, A4, PORTRAIT


class WkhtmltopdfConverter(HTMLConverter):
    """
    Implementation of HTML to PDF conversion using wkhtmltopdf tool.
    """

    def htmlconv(self, block, output_file, header_block=None, header_spacing=None,
                 footer_block=None, footer_spacing=None,  pdf_zoom=1, pdf_page_size=A4, orientation=PORTRAIT,
                 pdf_auto_shrink=True, **kwargs):
        """
        Use the wkhtmltohtml tool to convert the supplied HTML content to a PDF file.

        :param block: The block to render
        :param output_file: File name for converter output.
        :param header_block: A header block to add
        :param header_spacing: The spacing for the header as number in mm.
        :param footer_block: A footer block to add
        :param footer_spacing: The spacing for the footer as number in mm.
        :param pdf_zoom: The zooming to apply when rendering the page.
        :param pdf_page_size: The page size to use when rendering the page to PDF.
        :param orientation: Either html_converter.PORTRAIT or html_converter.LANDSCAPE
        :param kwargs: Additional parameters. Passed to wkhtmltopdf.
        """
        cmd = []
        cmd.append(self.get_executable('wkhtmltopdf'))
        cmd += ["--page-size", pdf_page_size]
        cmd += ["--orientation", orientation]
        cmd += ["--zoom", str(pdf_zoom)]
        if pdf_auto_shrink:
            cmd.append("--enable-smart-shrinking")
        else:
            cmd.append("--disable-smart-shrinking")

        # Specific wkhtmltox
        cmd += ["--no-stop-slow-scripts", "--debug-javascript"]
        cmd += ["--javascript-delay", str(kwargs.get('javascript_delay', 200))]
        [cmd.extend([k, v]) for k, v in kwargs.items()]

        # Render without header and footer as those are handles explicitly below
        content = block.render_html(static_output=True)
        html_filename = HTMLConverter.write_html_to_tempfile(block, content)

        temp_files = [html_filename]

        if header_block is not None:
            header_content = header_block.render_html(static_output=True)
            header_filename = HTMLConverter.write_html_to_tempfile(header_block, header_content)
            cmd += ['--header-html', header_filename]
            cmd += ['--header-spacing', str(header_spacing)]
            temp_files.append(header_filename)

        if footer_block is not None:
            footer_content = footer_block.render_html(static_output=True)
            footer_filename = HTMLConverter.write_html_to_tempfile(footer_block, footer_content)
            cmd += ['--footer-html', footer_filename]
            cmd += ['--footer-spacing', str(footer_spacing)]
            temp_files.append(footer_filename)

        cmd.extend([html_filename, output_file])

        output, errors = self.run_command(cmd)
        HTMLConverter.remove_temporary_files(temp_files)
        return output, errors


class WkhtmltoimageConverter(HTMLConverter):

    def htmlconv(self, block, output_file, pdf_zoom=1, **kwargs):
        """
        Use the wkhtmltoimage tool to convert the supplied HTML content to an image file.

        :param block: The block to render
        :param output_file: File name for converter output.
        :param pdf_zoom: The zooming to apply when rendering the page.
        :param kwargs: Additional parameters. Passed to wkhtmltoimage.
        """
        content = block.render_html(static_output=True)
        html_filename = HTMLConverter.write_html_to_tempfile(block, content)

        cmd = []
        cmd.append(self.get_executable('wkhtmltoimage'))
        cmd += ["--format", os.path.splitext(output_file)[1][1:]]
        cmd += ["--zoom", str(pdf_zoom)]
        # Set width to 0 to make it fit the content exactly, without any extra margin.
        cmd += ["--width", "0"]

        # Specific wkhtmltox
        cmd += ["--no-stop-slow-scripts", "--debug-javascript"]
        cmd += ["--javascript-delay", str(kwargs.get('javascript_delay', 200))]

        # Ignore default parameters that are not relevant for image output and pass the rest to wkhtmltoimage
        kwargs_ignore_list = ['header_block', 'header_spacing', 'footer_block', 'footer_spacing', 'pdf_auto_shrink',
                              'pdf_page_size', 'orientation']
        [cmd.extend([k, v]) for k, v in kwargs.items() if k not in kwargs_ignore_list]

        cmd.extend([html_filename, output_file])

        output, errors = self.run_command(cmd)
        HTMLConverter.remove_temporary_files([html_filename])
        return output, errors



from pybloqs.util import Cfg

from pybloqs.block.base import BaseBlock, HRule
from pybloqs.block.text import Raw, Pre, Span, Markdown
from pybloqs.block.layout import Grid, Flow, HStack, VStack
from pybloqs.block.convenience import Block
from pybloqs.block.image import ImgBlock, PlotBlock, set_plot_format
from pybloqs.block.table import HTMLJinjaTableBlock
from pybloqs.block.wrap import Box, Paragraph


__all__ = [
    # Core PyBloqs
    "set_plot_format",
    "Block",
    "BaseBlock",
    "HRule",
    "Raw",
    "Grid",
    "Flow",
    "HStack",
    "VStack",
    "ImgBlock",
    "PlotBlock",
    "HTMLJinjaTableBlock",
    "Box",
    "Paragraph",
    "Pre",
    "Span",
    "Markdown",
    "Cfg",
]




#!/bin/env python
from distutils.dir_util import mkpath
from distutils.file_util import copy_file
import glob
import logging
import os
from setuptools import setup, find_packages, Command
from setuptools.command.install import install
from setuptools.command.test import test as TestCommand
import six
import sys


# Convert Markdown to RST for PyPI
# http://stackoverflow.com/a/26737672
try:
    import pypandoc
    long_description = pypandoc.convert('README.md', 'rst')
    changelog = pypandoc.convert('CHANGES.md', 'rst')
except (IOError, ImportError, OSError):
    long_description = open('README.md').read()
    changelog = open('CHANGES.md').read()


def _copy_hc_files(source_paths, dest_path):
    """Copies all .js files (excluding .src.js) from source_path into dest_path."""
    if source_paths is None:
        logging.warn("***** Option --highcharts not specified. Highcharts blocks will not work! ******")
        return

    hc_files = []
    for source_path in source_paths:
        hc_files += set([i for i in glob.glob(os.path.join(source_path, '*.js')) if not i.endswith('.src.js')])
    if len(hc_files) > 0:
        for f in hc_files:
            copy_file(f, dest_path)
    else:
        logging.error(" No *.js files (excluding *.src.js) found in highcharts-paths: {}"
                      .format(source_paths))
        sys.exit(1)


def _copy_wkhtmltopdf(src_path):
    src_path = os.path.abspath(os.path.expanduser(src_path))
    files = ['wkhtmltopdf', 'wkhtmltoimage']
    if os.name == 'nt':
        files = [f + '.exe' for f in files]
    for f in files:
        source = os.path.join(src_path, f)
        dest = os.path.join(os.path.split(sys.executable)[0], f)
        copy_file(source, dest)


class LoadHighcharts(Command):
    user_options = [
        ("highcharts=", None, "List of paths with highcharts, e.g. ~/highcharts/js/,~/highcharts-heatmap/js/"),
    ]

    def initialize_options(self):
        self.highcharts = None

    def finalize_options(self):
        assert self.highcharts is not None, "Please provide --highcharts parameter"
        self.ensure_string_list("highcharts")
        self.highcharts = [os.path.abspath(os.path.expanduser(p)) for p in self.highcharts]

    def copy_hc_files(self):
        dest_path = os.path.join("pybloqs", "static")
        _copy_hc_files(self.highcharts, dest_path)

    def run(self):
        self.copy_hc_files()


class LoadWkhtmltopdf(Command):
    user_options = [
        ("wkhtmltopdf=", None, "Path for wkhtmltopdf and wkhtmltoimage."),
    ]

    def initialize_options(self):
        self.wkhtmltopdf = None

    def finalize_options(self):
        assert self.wkhtmltopdf is not None, "Please provide --wkhtmltopdf parameter"
        self.ensure_string("wkhtmltopdf")

    def run(self):
        _copy_wkhtmltopdf(self.wkhtmltopdf)


class PyBloqsInstall(install):
    # Options tuples: long name, short name and help string
    user_options = install.user_options + [
        ("highcharts=", None, "List of paths with highcharts, e.g. ~/highcharts/js/,~/highcharts-heatmap/js/"),
        ("wkhtmltopdf=", None, "Path for wkhtmltopdf and wkhtmltoimage."),
    ]

    def initialize_options(self):
        install.initialize_options(self)
        self.highcharts = None
        self.wkhtmltopdf = None

    def finalize_options(self):
        install.finalize_options(self)
        self.ensure_string_list("highcharts")
        if self.highcharts is not None:
            self.highcharts = [os.path.abspath(os.path.expanduser(p)) for p in self.highcharts]

    def copy_hc_files(self):
        dest_path = os.path.join(self.build_lib, "pybloqs", "static")
        _copy_hc_files(self.highcharts, dest_path)

    def minimise_js_files(self):
        """Find all .js files (including sub-directories) and minimise them with jsmin."""
        from jsmin import jsmin  # Lazy load to give setup machinery a chance to download dependency
        for dir_name, _, files in os.walk(os.path.join("build", "lib", "pybloqs", "static")):
            for f in files:
                if f.lower().endswith(".js"):
                    file_name = os.path.join(dir_name, f)
                    logging.info("Minimizing file: {}".format(file_name))
                    with open(file_name, 'r') as f_js:
                        content = f_js.read()
                    with open(file_name, 'w') as f_js:
                        f_js.write(jsmin(content))

    def run(self):
        logging.getLogger().setLevel(logging.INFO)
        mkpath(os.path.join(self.build_lib, 'pybloqs', 'static'))
        self.copy_hc_files()
        if self.wkhtmltopdf is not None:
            _copy_wkhtmltopdf(self.wkhtmltopdf)
        self.minimise_js_files()
        install.run(self)


class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = []

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        logging.basicConfig(format='%(asctime)s %(levelname)s %(name)s %(message)s', level='DEBUG')

        # import here, cause outside the eggs aren't loaded
        import pytest

        args = [self.pytest_args] if isinstance(self.pytest_args, six.string_types) else list(self.pytest_args)
        args.extend(['--cov', 'pybloqs',
                     '--cov-report', 'xml',
                     '--cov-report', 'html',
                     '--junitxml', 'junit.xml',
                     ])
        errno = pytest.main(args)
        sys.exit(errno)


setup(
    name="pybloqs",
    version="1.2.10",
    author="Man AHL Technology",
    author_email="ManAHLTech@ahl.com",
    description="Data Visualization and Report Building",
    long_description='\n'.join((long_description, changelog)),
    keywords=["ahl", "pdf", "html", "visualization", "report"],
    url="https://github.com/manahl/pybloqs",
    setup_requires=["jsmin"],
    install_requires=[
        "beautifulsoup4",
        "matplotlib",
        "markdown",
        "html5lib",
        "pandas",
        "docutils",
        "lxml",
        "pyyaml",
        "jinja2",
    ],
    extras_require={
        "docs_and_notebook": [
            "sphinx",
            "nbsphinx",
            "ipython[notebook]",
        ],
        "plotly": [
            "plotly"
        ],
        "bokeh": [
            "bokeh"
        ]
    },
    tests_require=[
        "mock",
        "pytest==4.6.4",
        "pytest-cov",
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Operating System :: OS Independent",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python",
        "Topic :: Scientific/Engineering",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.6",
    ],
    cmdclass={
        "install": PyBloqsInstall,
        "load_highcharts": LoadHighcharts,
        "load_wkhtmltopdf": LoadWkhtmltopdf,
        "test": PyTest,
    },
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    package_data={"pybloqs.static": ["*.js",
                                     "css/*.css",
                                     "css/pybloqs_default/main.css"],
                  "pybloqs.jinja": ["table.html"],
                  "pybloqs.htmlconv": ["*.js"]}
)



from pybloqs.block.table import HTMLJinjaTableBlock
from pybloqs.block.table_formatters import (fmt_fontsize_12,
                                            fmt_decimals_2,
                                            fmt_page_break,
                                            fmt_table_center,
                                            fmt_align_cells,
                                            fmt_align_header_index,
                                            fmt_header_index_bold,
                                            TableFormatter)
from pybloqs.static import JScript, Css


class DataTablesCSSClass(TableFormatter):

    def __init__(self, paging=True, searching=True, info=True):
        super(DataTablesCSSClass, self).__init__()
        self.paging = paging
        self.searching = searching
        self.info = info

    def create_table_level_css_class(self):
        no_paging = ' dt-no-paging' if self.paging is False else ''
        no_info = ' dt-no-info' if self.info is False else ''
        no_searching = ' dt-no-searching' if self.searching is False else ''
        css = 'blox_table compact row-border order-col stripe nowrap{}{}{}'.format(no_paging, no_info, no_searching)
        return css


class DataTablesHTMLJinjaTableBlock(HTMLJinjaTableBlock):

    resource_deps = [Css("css/jquery-dataTables"),
                     JScript("jquery"),
                     JScript("jquery-dataTables"),
                     JScript("jquery-dataTables-impl")]

    def __init__(self, df, formatters=None,
                 use_default_formatters=False,
                 paging=True,
                 searching=True,
                 info=True,
                 **kwargs):
        if formatters is None and use_default_formatters is False:
            formatters = [DataTablesCSSClass(paging, searching, info),
                          fmt_fontsize_12,
                          fmt_table_center,
                          fmt_align_cells,
                          fmt_align_header_index,
                          fmt_header_index_bold,
                          fmt_page_break,
                          fmt_decimals_2]
        super(DataTablesHTMLJinjaTableBlock, self).__init__(df, formatters, use_default_formatters, **kwargs)



import getpass
import os
import tempfile
import yaml

import pybloqs.util as util

# Default values that can be overridden by those stored in ~/.pybloqs.cfg
user_config = {
    "user_email_address": getpass.getuser(),
    "public_dir": "/tmp",
    "tmp_html_dir": tempfile.gettempdir(),
    "remove_temp_files": True,
    "smtp_kwargs": {
        "host": ""
    },
    "pdf_converter": "wkhtmltopdf",  # options: wkhtmltopdf or chrome_headless
    "image_converter": "wkhtmltoimage",
    "id_precision": 10,  # Number of digits to use from the id hash
}

try:
    stored_config = yaml.load(open(os.path.expanduser("~/.pybloqs.cfg"), "r"))
    if stored_config is not None:
        user_config.update(stored_config)
except IOError:
    pass
